@@ -5,6 +5,7 @@ import pandas as pd
 import numpy as np
 import logging
 import getpass
+import json
 from askdata.insight import Insight
 from askdata.channel import Channel
 from askdata.catalog import Catalog
@@ -172,41 +173,6 @@ class Agent(Insight, Channel, Catalog, Dataset):
         self._get_info_dataset_by_slug(slug)
         return self
 
-    def update_dataset_name(self, dataset_slug, dataset_name):
-
-        body = {"name": dataset_name}
-
-        s = requests.Session()
-        s.keep_alive = False
-        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
-        s.mount('https://', HTTPAdapter(max_retries=retries))
-
-        authentication_url = self._base_url_askdata + '/smartdataset/datasets/' + dataset_slug + '/sdk'
-        logging.info("AUTH URL {}".format(authentication_url))
-
-        headers = {
-            "Content-Type": "application/json",
-            "Authorization": "Bearer" + " " + self._token
-        }
-        response = s.put(url=authentication_url, json=body, headers=headers)
-        response.raise_for_status()
-
-    def create_parquet_dataset(self, agent_slug, dataset_slug, file_path):
-
-        s = requests.Session()
-        s.keep_alive = False
-        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
-        s.mount('https://', HTTPAdapter(max_retries=retries))
-
-        authentication_url = self._base_url_askdata + '/smartbot/agents/' + agent_slug + '/datasets/' + dataset_slug + '/parquet'
-        logging.info("AUTH URL {}".format(authentication_url))
-        file = {'file': open(file_path, 'rb')}
-        headers = {"Authorization": "Bearer" + " " + self._token}
-        response = s.post(url=authentication_url, files=file, headers=headers)
-        response.raise_for_status()
-        r = response.json()
-
-
     def query(self, dataset_slug="", fields:list=[], filters=None, sorted_by=None,  dataset_id="", limit=50):
 
         if(fields==[]):
@@ -282,6 +248,40 @@ class Agent(Insight, Channel, Catalog, Dataset):
         return pd.DataFrame(r['data'])
 
 
+    def update_dataset_name(self, dataset_slug, dataset_name):
+
+        body = {"name": dataset_name}
+
+        s = requests.Session()
+        s.keep_alive = False
+        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
+        s.mount('https://', HTTPAdapter(max_retries=retries))
+
+        authentication_url = self._base_url_askdata + '/smartdataset/datasets/' + dataset_slug + '/sdk'
+        logging.info("AUTH URL {}".format(authentication_url))
+
+        headers = {
+            "Content-Type": "application/json",
+            "Authorization": "Bearer" + " " + self._token
+        }
+        response = s.put(url=authentication_url, json=body, headers=headers)
+        response.raise_for_status()
+
+    def create_parquet_dataset(self, agent_slug, dataset_slug, file_path):
+
+        s = requests.Session()
+        s.keep_alive = False
+        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
+        s.mount('https://', HTTPAdapter(max_retries=retries))
+
+        authentication_url = self._base_url_askdata + '/smartbot/agents/' + agent_slug + '/datasets/' + dataset_slug + '/parquet'
+        logging.info("AUTH URL {}".format(authentication_url))
+        file = {'file': open(file_path, 'rb')}
+        headers = {"Authorization": "Bearer" + " " + self._token}
+        response = s.post(url=authentication_url, files=file, headers=headers)
+        response.raise_for_status()
+        r = response.json()
+
 
     def update_parquet_dataset(self, agent_slug, dataset_id, file_path, strategy):
 
@@ -353,8 +353,6 @@ class Agent(Insight, Channel, Catalog, Dataset):
         dataset_id = dataset_id
         url_get = self.smart_insight_url + "/composed_queries?datasetId=" + dataset_id
 
-        logging.info("CALLING {}".format(url_get))
-
         s = requests.Session()
         s.keep_alive = False
         retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
@@ -411,7 +409,7 @@ class Agent(Insight, Channel, Catalog, Dataset):
         r = response.json()
         return pd.DataFrame(r['data'])
 
-    def get_dataset_by_slug(self, agent_slug, slug: str) -> Dataset:
+    def get_dataset_by_slug(self, agent_slug, slug: str):
         s = requests.Session()
         s.keep_alive = False
         retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
@@ -429,7 +427,7 @@ class Agent(Insight, Channel, Catalog, Dataset):
         except:
             return None
 
-    def _update_dataset_icon(self, dataset_id, icon_url):
+    def __update_dataset_icon(self, dataset_id, icon_url):
 
         url = self._base_url_askdata + "/smartdataset/datasets/" + dataset_id + "/settings"
 
@@ -502,7 +500,7 @@ class Agent(Insight, Channel, Catalog, Dataset):
         dataset_id = response.json()["id"]
 
         if (icon_url != None):
-            self._update_dataset_icon(dataset_id, icon_url)
+            self.__update_dataset_icon(dataset_id, icon_url)
 
     def update_dataset(self, dataframe: pd.DataFrame, dataset_name: str, slug: str, icon_url=None,
                        settings: dict = None):
@@ -540,29 +538,8 @@ class Agent(Insight, Channel, Catalog, Dataset):
         dataset_id = response.json()["id"]
 
         if (icon_url != None):
-            self._update_dataset_icon(dataset_id, icon_url)
+            self.__update_dataset_icon(dataset_id, icon_url)
 
-    '''La prima chiave è la stringa colonna del dataframe. Per ogni chiave ho dei setting:
-    name stringa
-    type(dimension o measure)
-    isDate booleano
-    synonyms: [] array di stringhe
-    dateFormat(pattern della data)
-    numberFormat
-    defaultAggregation(sum, avg..)
-
-    {
-        "column_name": {
-            "name": "column_name",
-            "type": "Dimension",
-            "isDate": False,
-            "synonyms": [],
-            "dateFormat",
-            "numberFormat": "",
-            "defaultAggregation": "sum"
-        }
-    }
-    '''
 
     def check_dataset_columns(self, dataframe: pd.DataFrame, dataset_id):
 
@@ -590,7 +567,27 @@ class Agent(Insight, Channel, Catalog, Dataset):
                 logging.info("Column {} is being {}".format(differece["columnName"], differece["differenceType"]))
         return r["hasDifference"]
 
+    '''La prima chiave è la stringa colonna del dataframe. Per ogni chiave ho dei setting:
+    name stringa
+    type(dimension o measure)
+    isDate booleano
+    synonyms: [] array di stringhe
+    dateFormat(pattern della data)
+    numberFormat
+    defaultAggregation(sum, avg..)
 
+    {
+        "column_name": {
+            "name": "column_name",
+            "type": "Dimension",
+            "isDate": False,
+            "synonyms": [],
+            "dateFormat",
+            "numberFormat": "",
+            "defaultAggregation": "sum"
+        }
+    }
+    '''
     def create_or_replace_dataset(self, dataframe: pd.DataFrame, dataset_name: str, slug: str, icon_url=None,
                                   settings: dict = None):
         # Check if dataset exists
@@ -615,7 +612,7 @@ class Agent(Insight, Channel, Catalog, Dataset):
             self._delete_dataset(dataset_id)
             logging.info("---- dataset '{}' deleted ----- ".format(dataset_id))
         else:
-            raise Exception('takes 2 positional arguments "slug, datset_id" but 0 were given')
+            raise Exception('Please give either a dataset id or a dataset slug')
 
     def create_datacard(self, channel_name: str, title: str, search: str = "", slug: str = None, replace: bool = False):
 
@@ -629,7 +626,7 @@ class Agent(Insight, Channel, Catalog, Dataset):
             print("Please specify a datacard slug to replace!")
             return
 
-        channel = self.get_channel(self._agentId, channel_name)
+        channel = self.get_channel(channel_name)
         if channel != None:
             channel_id = channel["id"]
         else:
@@ -723,14 +720,14 @@ class Agent(Insight, Channel, Catalog, Dataset):
         r.raise_for_status()
         return r.json()['id']
 
-    def get_channel(self, agent_id, channel_code):
+    def get_channel(self, channel_code):
 
         s = requests.Session()
         s.keep_alive = False
         retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
         s.mount('https://', HTTPAdapter(max_retries=retries))
 
-        url = self._base_url_ch + '/channels?agentId=' + agent_id + '&slug=' + channel_code
+        url = self._base_url_ch + '/channels?agentId=' + self._agentId + '&slug=' + channel_code
         r = s.get(url=url, headers=self._headers)
         r.raise_for_status()
         if (r != None and r.json() != []):
@@ -766,17 +763,22 @@ class Askdata(SignUp):
 
             self._token = token
             self._domainlogin = domainlogin.upper()
-            self._env = env
+            self._env = env.lower()
 
             if self._env == 'dev':
                 self.base_url_security = url_list['BASE_URL_SECURITY_DEV']
-
-            if self._env == 'qa':
+            elif self._env == 'qa':
                 self.base_url_security = url_list['BASE_URL_SECURITY_QA']
-
-            if self._env == 'prod':
+            elif self._env == 'prod':
                 self.base_url_security = url_list['BASE_URL_SECURITY_PROD']
 
+            if self._env == 'dev':
+                self.base_url = url_list['BASE_URL_ASKDATA_DEV']
+            elif self._env == 'qa':
+                self.base_url = url_list['BASE_URL_ASKDATA_QA']
+            elif self._env == 'prod':
+                self.base_url = url_list['BASE_URL_ASKDATA_PROD']
+
             if token == '':
 
                 if username == '':
@@ -825,22 +827,11 @@ class Askdata(SignUp):
 
     def join(self, df:pd.DataFrame, with_query:str, workspace:str, how="inner", on=None, left_on=None, right_on=None, lang="en"):
 
-        if self._env == 'dev':
-            base_url = url_list['BASE_URL_ASKDATA_DEV']
-        elif self._env == 'qa':
-            base_url = url_list['BASE_URL_ASKDATA_QA']
-        elif self._env == 'prod':
-            base_url = url_list['BASE_URL_ASKDATA_PROD']
-
-        if(on==None and (left_on== None and right_on==None)):
-            logging.error("Pleanse insert either 'on' condition or right_on and left_on conditions!")
-            return
-
         s = requests.Session()
         s.keep_alive = False
         retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
         s.mount('https://', HTTPAdapter(max_retries=retries))
-        authentication_url = base_url + '/smartbot/agents/agentslug/' + workspace
+        authentication_url = self.base_url + '/smartbot/agents/agentslug/' + workspace
 
         headers = {"Authorization": "Bearer" + " " + self._token}
         response = s.get(url=authentication_url, headers=headers)
@@ -848,7 +839,7 @@ class Askdata(SignUp):
         r = response.json()
         agent_id = r["agentId"]
 
-        url = base_url + "/smartbot/agents/"+agent_id+"/ask"
+        url = self.base_url + "/smartbot/agents/"+agent_id+"/ask"
 
         s = requests.Session()
         s.keep_alive = False
@@ -866,17 +857,155 @@ class Askdata(SignUp):
         result = response.json()["result"]
         df_query = pd.DataFrame(data=result[0]["attachment"]["body"][0]["details"]["rows"], columns=result[0]["attachment"]["body"][0]["details"]["columns"])
 
+        if (on == None and (left_on == None and right_on == None)):
+            join_conditions = self.__get_join_info(df, df_query)["join_list"][0]
+            return df.merge(df_query, how, '', join_conditions["left_on"], join_conditions["right_on"])
+
         return df.merge(df_query, how, on, left_on, right_on)
 
+    def get_join_info(self, df1: pd.DataFrame, df2: pd.DataFrame):
+        return self.__get_join_info(df1, df2)
 
-    def get(self, query, workspace='', dataset='', lang="en"):
+    def __get_join_info(self, df1:pd.DataFrame, df2:pd.DataFrame):
+
+        columns = []
+        for field in json.loads(df1.to_json(orient="table", index=False))["schema"]["fields"]:
+            field["label"] = field.pop("name")
+            field["type"] = self.__choose_type(field.pop("type"))
+            columns.append(field)
+
+
+        columns2 = []
+        for field in json.loads(df2.to_json(orient="table", index=False))["schema"]["fields"]:
+            field["label"] = field.pop("name")
+            field["type"] = self.__choose_type(field.pop("type"))
+            columns2.append(field)
+
+
+        table_left = {
+            "columns": columns,
+            "values": json.loads(df1.to_json(orient="values"))[:5]
+        }
+        table_right = {
+            "columns": columns2,
+            "values": json.loads(df2.to_json(orient="values"))[:5]
+        }
+        tables = {
+            "tables": {
+                "table_left": table_left,
+                "table_right": table_right
+            }
+        }
+        print(tables)
+        s = requests.Session()
+        s.keep_alive = False
+        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
+        s.mount('https://', HTTPAdapter(max_retries=retries))
+
+        url = self.base_url + '/h2sql/smartjoin'
+
+        r = s.post(url=url, headers=self._headers, json=tables)
+        r.raise_for_status()
+
+        try:
+            return r.json()["join_list"][0]
+        except:
+            logging.error(r.json())
+
+    def __choose_type(self, var):
+        if "int" in var or "num" in var or "real" in var or "double" in var or "decimal" in var or "float" in var:
+            return "numeric"
+        elif "text" in var or "varchar" in var or "char" in var or "default" in var:
+            return "string"
+        elif "date" in var or "year" in var:
+            return "datetime"
+        elif "time" in var:
+            return "timestamp"
+        elif "bool" in var or "bit" in var:
+            return "bool"
+        elif "blob" in var:
+            return "blob"
+        else:
+            return var
+
+    def get_columns(self, dataset, workspace=''):
+
+        if(workspace==''):
+            if('/' in dataset):
+                workspace=dataset.split("/")[0]
+                slug = dataset.split("/")[1]
+        else:
+            slug=dataset
+
+        s = requests.Session()
+        s.keep_alive = False
+        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
+        s.mount('https://', HTTPAdapter(max_retries=retries))
+
+        url = self.base_url + '/smartbot/agents/agentslug/' + workspace + '/datasetslug/' + slug
+
+        headers = {"Authorization": "Bearer" + " " + self._token}
+        response = s.get(url=url, headers=headers)
+        response.raise_for_status()
+        r = response.json()
+        agent_id = r["agentId"]
+        dataset_id = r["datasetId"]
+
+        s = requests.Session()
+        s.keep_alive = False
+        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
+        s.mount('https://', HTTPAdapter(max_retries=retries))
+
+        url = self.base_url + '/smartdataset/datasets/' + dataset_id + '/measures'
+
+        headers = {"Authorization": "Bearer" + " " + self._token}
+        response = s.get(url=url, headers=headers)
+        response.raise_for_status()
+        measures = response.json()
+        columns=[]
+        for measure in measures:
+            columns.append({
+                "code": measure["code"],
+                "name": measure["name"],
+                "type": "MEASURE",
+                "internalType": measure["schemaMetaData"]["dataType"]
+            })
+        s = requests.Session()
+        s.keep_alive = False
+        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
+        s.mount('https://', HTTPAdapter(max_retries=retries))
+
+        url = self.base_url + '/smartdataset/datasets/' + dataset_id + '/entityTypes'
+
+        headers = {"Authorization": "Bearer" + " " + self._token}
+        response = s.get(url=url, headers=headers)
+        response.raise_for_status()
+        entityTypes = response.json()
+
+        for entityType in entityTypes:
+            columns.append({
+                "code": entityType["code"],
+                "name": entityType["name"],
+                "type": "ENTITY_TYPE",
+                "internalType": entityType["schemaMetaData"]["dataType"]
+            })
+        return pd.DataFrame(data=columns)
+
+
+    def query(self, workspace='', dataset='', fields:list=[], filters=None, sorted_by=None, limit=50):
+
+        if (workspace == ''):
+            if ('/' in dataset):
+                workspace = dataset.split("/")[0]
+                dataset_slug = dataset.split("/")[1]
+
+
+        agent = self.agent(slug=workspace)
+        return agent.query(dataset_slug, fields, filters, sorted_by, limit=limit)
+
+
+    def get(self, query, workspace='', dataset='', lang="en", fields:list=[], filters=None, sorted_by=None):
 
-        if self._env == 'dev':
-            base_url = url_list['BASE_URL_ASKDATA_DEV']
-        elif self._env == 'qa':
-            base_url = url_list['BASE_URL_ASKDATA_QA']
-        elif self._env == 'prod':
-            base_url = url_list['BASE_URL_ASKDATA_PROD']
 
         s = requests.Session()
         s.keep_alive = False
@@ -888,8 +1017,7 @@ class Askdata(SignUp):
         elif(workspace != '' and dataset==''):
             dataset_id = ''
             try:
-
-                authentication_url = base_url+ '/smartbot/agents/agentslug/' + workspace
+                authentication_url = self.base_url+ '/smartbot/agents/agentslug/' + workspace
 
                 headers = {"Authorization": "Bearer" + " " + self._token}
                 response = s.get(url=authentication_url, headers=headers)
@@ -907,7 +1035,7 @@ class Askdata(SignUp):
                 workspace=dataset.split("/")[0]
                 slug=dataset.split("/")[1]
 
-                authentication_url = base_url+'/smartbot/agents/agentslug/'+workspace + '/datasetslug/' + slug
+                authentication_url = self.base_url+'/smartbot/agents/agentslug/'+workspace + '/datasetslug/' + slug
 
                 headers = {"Authorization": "Bearer" + " " + self._token}
                 response = s.get(url=authentication_url, headers=headers)
@@ -921,7 +1049,7 @@ class Askdata(SignUp):
                 return None
 
 
-        url = base_url + "/smartbot/agents/"+agent_id+"/ask"
+        url = self.base_url + "/smartbot/agents/"+agent_id+"/v2/ask"
 
         s = requests.Session()
         s.keep_alive = False
@@ -931,32 +1059,45 @@ class Askdata(SignUp):
         body = {
             "message": query,
             "language": lang.lower(),
-            "userId": self.userid
+            "userId": self.userid,
+            "fields": fields,
+            "filters": filters,
+            "sorting": sorted_by
         }
 
         response = s.post(url=url, json=body, headers=self._headers)
         response.raise_for_status()
-        result = response.json()["result"]
+        result = response.json()
+        try:
+            if(dataset_id!=''):
+                filtered_result = []
+                for r in result:
+                    if(r["attachment"]["body"][0]["details"]["datasetId"]==dataset_id):
+                        filtered_result.append(r)
+            else:
+                filtered_result = result
 
-        if(dataset_id!=''):
-            filtered_result = []
-            for r in result:
-                if(r["attachment"]["body"][0]["details"]["datasetId"]==dataset_id):
-                    filtered_result.append(r)
-        else:
-            filtered_result = result
+            if self._env == 'dev':
+                app_url = url_list['BASE_URL_APP_DEV']
+            elif self._env == 'qa':
+                app_url = url_list['BASE_URL_APP_QA']
+            elif self._env == 'prod':
+                app_url = url_list['BASE_URL_APP_PROD']
 
-        #print("{} from the dataset: {} / {}".format(self.extract_fields(result[0]["attachment"]["body"][0]["details"]["columns"]), workspace, result[0]["attachment"]["body"][0]["details"]["datasetName"]))
-        display(HTML("<p>{} from the dataset: {} / {}</p>".format(self.extract_fields(filtered_result[0]["attachment"]["body"][0]["details"]["columns"]), workspace, filtered_result[0]["attachment"]["body"][0]["details"]["datasetName"])))
-        if(len(filtered_result[0]["attachment"]["body"])>1):
-            display(HTML("<p>Filtered by: {}</p>".format(filtered_result[0]["attachment"]["body"][1]["details"]["text"])))
-        if(len(filtered_result)>1):
-            display(HTML("<p>{} Alternatives:</p>".format(len(filtered_result)-1)))
-            for r in filtered_result[1:]:
-                display(HTML("<p>     {} from the dataset: {} / {}</p>".format(self.extract_fields(r["attachment"]["body"][0]["details"]["columns"]), workspace, r["attachment"]["body"][0]["details"]["datasetName"])))
+            url_1 = app_url + "/" + workspace+"/"+filtered_result[0]["attachment"]["body"][0]["details"]["datasetSlug"]
+            display(HTML('<p>{} from the dataset: {} / {} : (<a href="{}" target="_blank">{}</a>) <- Use this source</p>'.format(self.extract_fields(filtered_result[0]["attachment"]["body"][0]["details"]["columns"]), workspace, filtered_result[0]["attachment"]["body"][0]["details"]["datasetName"], url_1, url_1)))
+            if(len(filtered_result[0]["attachment"]["body"])>1):
+                display(HTML("<p>Filtered by: {}</p>".format(filtered_result[0]["attachment"]["body"][1]["details"]["text"])))
+            if(len(filtered_result)>1):
+                display(HTML("<p>{} Alternatives:</p>".format(len(filtered_result)-1)))
+                for r in filtered_result[1:]:
+                    url_i = app_url + "/" + workspace + "/" + r["attachment"]["body"][0]["details"]["datasetSlug"]
+                    display(HTML('<p>     {} from the dataset: {} / {} ((<a href="{}" target="_blank">{}</a>)) <- Use this source</p>'.format(self.extract_fields(r["attachment"]["body"][0]["details"]["columns"]), workspace, r["attachment"]["body"][0]["details"]["datasetName"], url_i, url_i)))
 
-        df = pd.DataFrame(data=filtered_result[0]["attachment"]["body"][0]["details"]["rows"], columns=filtered_result[0]["attachment"]["body"][0]["details"]["columns"])
-        return df
+            df = pd.DataFrame(data=filtered_result[0]["attachment"]["body"][0]["details"]["rows"], columns=filtered_result[0]["attachment"]["body"][0]["details"]["columns"])
+            return df
+        except:
+            logging.error("Query did not produce any result")
 
     def load_agents(self):
 
@@ -987,10 +1128,6 @@ class Askdata(SignUp):
         response = super().signup_user(username, password, firstname, secondname, title)
         return response
 
-    @property
-    # ?
-    def responce(self):
-        return self.r2
 
     def create_agent(self, agent_name):
 
@@ -999,18 +1136,11 @@ class Askdata(SignUp):
             "language": "en"
         }
 
-        if self._env == 'dev':
-            self._base_url = url_list['BASE_URL_ASKDATA_DEV']
-        if self._env == 'qa':
-            self._base_url = url_list['BASE_URL_ASKDATA_QA']
-        if self._env == 'prod':
-            self._base_url = url_list['BASE_URL_ASKDATA_PROD']
-
         s = requests.Session()
         s.keep_alive = False
         retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
         s.mount('https://', HTTPAdapter(max_retries=retries))
-        authentication_url = self._base_url + '/smartbot/agents'
+        authentication_url = self.base_url + '/smartbot/agents'
         r = s.post(url=authentication_url, headers=self._headers, json=data)
         r.raise_for_status()
 
